#ifndef	_MLFdatum_h_
#define	_MLFdatum_h_

#include <Rtypes.h>

#include <arpa/inet.h>

#include "module_numbers.h"

class MLFdatum {
private:
    union {
	UChar_t fBuf[8];
	ULong64_t fVal;
	UInt_t  fVal32[2];
	struct {
	    UChar_t fHeader; /* 0x22 TDChit, 0x20 comment, 0x5b/0x5c GateNet */
	    UChar_t fReserved;
	    UShort_t fModule;
	    UShort_t fCh;
	    UShort_t fTime;
	} m_FPGA_TDC; /* COPPER, COPPER-Lite, Kalliope, NIM-TDC */
	struct {
	    UChar_t fHeader; /* 0x42 TDC 0x44 ADC */
	    UChar_t fResolution; /* only lower 4bits are used */
	    UChar_t fModuleCh; /* Higher 4bits for module, lower 4bits for channel */
	    UChar_t fTimeHi;
	    UInt_t fTimeLo;
	} m_NIKI; /* Niki 3n00 */
    };

public:

    void Set64(ULong64_t val) {
	fVal = val;
    };

    void Set32(UInt_t val) {
	fVal = 0;
	m_FPGA_TDC.fHeader = 0x20;
	fVal32[1] = val;
    };

    void Set16(UShort_t val) {
	fVal = 0;
	m_FPGA_TDC.fHeader = 0x20;
	m_FPGA_TDC.fTime = htons(val);
    };

    void Set32N(UInt_t val, UChar_t header) {
	fVal = 0;
	m_FPGA_TDC.fHeader = header;
	fVal32[1] = val;
    };

    void Set16N(UShort_t val, UChar_t header) {
	fVal = 0;
	m_FPGA_TDC.fHeader = header;
	m_FPGA_TDC.fTime = htons(val);
    };

    void SetTDC(UShort_t mod, UShort_t ch, UShort_t time) {
	fVal = 0;
	m_FPGA_TDC.fHeader = 0x22;
	m_FPGA_TDC.fModule = htons(mod);
	m_FPGA_TDC.fCh = htons(ch);
	m_FPGA_TDC.fTime = htons(time);
	
    };

    ULong64_t GetVal() const {
	return fVal;
    };
    UChar_t * GetBuf() {
	return fBuf;
    };

    Int_t GetType() const { return fBuf[0]; };

    void SetType(Int_t type) {
	fBuf[0] = type;
    };

    Int_t digit_to_ns() const {
	static const int mag[] = { 5, 10, 20, 100, 200, 500, 1000 };
	return mag[m_NIKI.fResolution & 7];
    }

    Long64_t GetTime() const {
	Long64_t digit;
	switch (GetType()) {
	case 0x10:
	case 0x11:
	case 0x20:
	case 0x22:
	    return ntohs(m_FPGA_TDC.fTime);
	case 0x40:
	case 0x42:
	  digit = (((Long64_t)m_NIKI.fTimeHi << 32) + ntohl(m_NIKI.fTimeLo));
	    return (digit * digit_to_ns());
	case 0x46:
	  digit = (((Long64_t)m_NIKI.fTimeHi << 16) + (ntohl(m_NIKI.fTimeLo)>>16));
	    return (digit * digit_to_ns());
	default:
	    return -1;
	}
    };

    Int_t GetADC() const {
	/* ADC is 16bit, so Int_t is sufficient */
	switch (GetType()) {
	  //	case 0x42:
	case 0x40:
	case 0x44:
	case 0x46:
	  return ntohl(m_NIKI.fTimeLo) & 0x0000FFFF;
	default:
	    return -1;
	}
    };

    UInt_t GetCh() const {
	switch (GetType()) {
	case 0x10:
	case 0x20:
	case 0x40:
	case 0x22:
	    return ntohs(m_FPGA_TDC.fCh); /* short */
	case 0x42: /* Niki TDC */
	case 0x44: /* Niki ADC */
	case 0x46: /* Niki TDCADC */
	    return (m_NIKI.fModuleCh & 0xF); /* char */
	}
	return 0;
    };

    Int_t GetModule() const {
	switch (GetType()) {
	case 0x22:
	  return ntohs(m_FPGA_TDC.fModule);
	case 0x42: /* Niki TDC */
	case 0x44: /* Niki ADC */
	case 0x46: /* Niki TDCADC */
	    return (m_NIKI.fModuleCh >> 4); /* char */
	default:
	    return -1;
	}
    };

    /* Niki */

    void SetNikiTDCADC(int module, int ch, int resolution, Long64_t tdc, UInt_t adc) {
	m_NIKI.fHeader = 0x46; /* This is Niki TDCADC */
	m_NIKI.fResolution = resolution;
	m_NIKI.fModuleCh = (UChar_t)(((module & 0xF) << 4) | (ch & 0xF));
	m_NIKI.fTimeHi = ((tdc>>16)&0xFF);
	//	printf("adc: %x %ld\n",m_NIKI.fHeader,adc);;
	m_NIKI.fTimeLo = htonl(((tdc&0x0000FFFF)<<16)+adc);
    };

    void SetNikiADC(int module, int ch, UInt_t adc) {
	m_NIKI.fHeader = 0x44; /* This is Niki ADC */
	m_NIKI.fResolution = 0;
	m_NIKI.fModuleCh = (UChar_t)(((module & 0xF) << 4) | (ch & 0xF));
	m_NIKI.fTimeHi = 0;
	//	printf("adc: %x %ld\n",m_NIKI.fHeader,adc);;
	m_NIKI.fTimeLo = htonl(adc);
    };

    void SetNikiTDC(int module, int ch, int resolution, Long64_t tdc) {
	m_NIKI.fHeader = 0x42; /* This is Niki TDC */
	m_NIKI.fResolution = resolution;
	m_NIKI.fModuleCh = (((module & 0xF) << 4) | (ch & 0xF));
	//	printf("tdc: %ld",tdc&0xFFFFFFFF);;
	m_NIKI.fTimeHi = ((tdc >> 32) & 0xFF);
	m_NIKI.fTimeLo = htonl(tdc & 0xFFFFFFFF);
    };

    Int_t GetResolution() const {
	if (m_NIKI.fHeader == 0x42) {
	    return m_NIKI.fResolution;
	} else {
	    return 1; /* 1ns from COPPER-Lite, Kalliope */
	}
    };

};
#endif
