#include "module_numbers.h"
#include "Run44Decoder.h"
#include "MLFdatum.h"
#define POSITRON1 7600
#define POSITRON2 7760
#define MUON1 8620
#define MUON2 40000

ClassImp(Run44Decoder)

//Decode one event for one trigger pulse
void Run44Decoder::Push(MLFdatum & item, int adc_info) {
    fNhit_single_per_100++;
    fNhit_single++;

#if 0
    /* debug purpose */
    UChar_t * p = item.GetBuf();
    for (int i=0; i<8; i++) {
	printf(" %02x", p[i]);
    }
    putchar('\n');
#endif

    Int_t type = item.GetType();
    Int_t ch = item.GetCh();               //channel of individual modules
    Int_t module_id = item.GetModule();  //FOR TEST OF +1: NIM-TDC +2: NIKI MODULES
    Int_t time = item.GetTime();
    Int_t height = adc_info;
    Int_t fRG=GetRG();

    if(fRG >= MAX_RG){
      printf("\nNIM-TDC input register pattern 0..%d exceeds RG slot prepared: MAX_RG=%d\n",fRG,MAX_RG);
      exit(1);
    }
    //    if(module_id){
    //    printf("item: %0X",item.GetType());
    //      printf("ch; %ld",ch);
    //      printf("time: %ld",time);
    //      printf("height: %ld\n",height);
      //    }
    //    cerr<<module_id<<" "<<ch<<endl;
    //calculate module_id X ch -> histogram channels

    int Base,Top,Channel,BaseCh,bare_ch;

    switch(type){
    case 0x22: 
      //    if(type==0x22){
      Base=0;BaseCh=0;
      Top=Base+N_COPPER_LITE_MODULES;
      Channel=N_COPPER_LITE_CHANNELS;
      if(module_id>=Base && module_id<Top){
	bare_ch=ch;
	ch += (module_id-Base)*Channel;
      }
      Base=Top;
      BaseCh+=N_COPPER_LITE_MODULES*N_COPPER_LITE_CHANNELS;
      Top=Base+N_KALLIOPE_TDC_MODULES;
      Channel=N_KALLIOPE_TDC_CHANNELS;
      if(module_id>=Base && module_id<Top){
	bare_ch=ch;
	ch += (module_id-Base)*Channel+BaseCh;
      }
      Base=Top;
      BaseCh+=N_KALLIOPE_TDC_MODULES*N_KALLIOPE_TDC_CHANNELS;
      Top=Base+N_NIM_TDC_MODULES;
      Channel=N_NIM_TDC_CHANNELS;
      if(module_id>=Base && module_id<Top){
	bare_ch=ch;
	ch += (module_id-Base)*Channel+BaseCh;
      }
      break;
      
      //    }else if(type==0x42 || type==0x44){
	//    }else if(type==0x46){
    case 0x42:
    case 0x44:
    case 0x46:
      Base=N_COPPER_LITE_MODULES+N_KALLIOPE_TDC_MODULES+N_NIM_TDC_MODULES;
      BaseCh=N_COPPER_LITE_MODULES*N_COPPER_LITE_CHANNELS
	+N_KALLIOPE_TDC_MODULES*N_KALLIOPE_TDC_CHANNELS+N_NIM_TDC_MODULES*N_NIM_TDC_CHANNELS;
      Top=Base+N_NIKI_3N00_MODULES;
      Channel=N_NIKI_3N00_CHANNELS;
      //      if(module_id>=Base && module_id<Top){
      //      printf("%d %d: %d %d %d %d\n",module_id,ch,Base, Top, BaseCh,Channel);
      //	ch = ch*2+(module_id-Base)*Channel+BaseCh;
      bare_ch=ch;
      ch =ch*2+(module_id-1)*Channel+BaseCh;
      //	if(type==0x44) ch++;
      //      }
      break;
    default: 
      fprintf(stderr,"undermined ch %d with type %02x\n",ch,type);
      return;
    }

    	if (ch >= fNch) {
	  fprintf(stderr, "************************\n Type: %0x ch (%d x 32 + %d) = %d > fNch %d \n************************\n",item.GetType(), item.GetModule(), item.GetCh(), ch, fNch);
	}

    if ( time >= POSITRON1 && time <= POSITRON2 ) fNhit_positron_per_100++;
    if ( time >= MUON1 && time <= MUON2 ) fNhit_muon_per_100++;

    TString temp;
    temp=GetCounterMap(ch);
    if(strpbrk(temp.Data(),"F")) fNhit_singleFWD_per_100++;
    if(strpbrk(temp.Data(),"B")) fNhit_singleBWD_per_100++;
    if(strpbrk(temp.Data(),"L")) fNhit_singleLFT_per_100++;
    if(strpbrk(temp.Data(),"R")) fNhit_singleRIT_per_100++;
    if(strpbrk(temp.Data(),"U")) fNhit_singleUP_per_100++;
    if(strpbrk(temp.Data(),"D")) fNhit_singleDWN_per_100++;

    if(IsKalliope(ch)||IsCopperLite(ch)){
      fH2Any.Fill(time, ch); //filling 2D histogram map
      fH2single.Fill(time, ch);//filling 2D histogram map
      fSingleHitList[ch].Add(new Hit(ch, time));
      fSingleHits.Add(new Hit(ch, time));

      fAnyHitList[ch].Add(new Hit(ch, time));
      fAnyHits.Add(new Hit(ch, time));

      if (GetSingleHist(ch+fNch*fRG) != NULL) {
// cout << "single " << ch << " " << time << "ns" << endl;
	GetSingleHist(ch+fNch*fRG)->Fill(time);//filling bare histogram
      } else {
	printf("GetSingleHist(%d) returned NULL\n", ch);
      }
      
      Fill(&fSingleGroupList, ch, time, 0, 0);
      Fill(&fSingleAsymmetryList, ch, time, 0, 0);
    }

    if(IsNiki(ch)){
      switch(type){
      case 0x46: 
      case 0x42:
	//	printf("%d: %d %d %d\n",ch, type,time,height);
        fH2TDCADC.Fill(time, height); //filling TDC vs. ADC map
	fAnyHitList[ch+1].Add(new Hit(ch+1, height));
        fH2Any.Fill(height, ch+1);    //filling ADC vs. ch ma
	fAnyHits.Add(new Hit(ch+1, height));
        fH2ADC.Fill(height, bare_ch);//filling ADC vs.ch map by module
	if (GetADCHist(ch+1+fNch*fRG) != NULL) {
	  GetADCHist(ch+1+fNch*fRG)->Fill(height);//filling bare histogram
	} else {
	  printf("GetADCHist(%d) returned NULL\n", ch);
	}
	Fill(&fADCGroupList, ch+1, height, 0, 0);    
	Fill(&fADCCutGroupList, ch+1, height, ch, time);//***

	fAnyHitList[ch].Add(new Hit(ch, time));
        fH2Any.Fill(time, ch);        //filling TDC vs. ch map
	fAnyHits.Add(new Hit(ch, time));
	fH2TDC.Fill(time,   bare_ch);//filling TDC vs.ch map by module
	if (GetTDCHist(ch+fNch*fRG) != NULL) {
	  GetTDCHist(ch+fNch*fRG)->Fill(time);//filling bare histogram
	} else {
	  printf("GetTDCHist(%d) returned NULL\n", ch);
	}
	Fill(&fTDCGroupList, ch, time, 0, 0);
	Fill(&fTDCCutGroupList, ch, time, ch+1, height);//***

      case 0x44: break;
      }
    }

    if(IsNim(ch)){
      fH2Any.Fill(time, ch); //filling 2D histogram map
      fH2NDC.Fill(time, bare_ch);//filling 2D histogram map
      fAnyHitList[ch].Add(new Hit(ch, time));
      fAnyHits.Add(new Hit(ch, time));
      if (GetNDCHist(ch+fNch*fRG) != NULL) {
	// cout << "single " << ch << " " << time << "ns" << endl;
	GetNDCHist(ch+fNch*fRG)->Fill(time); //filling bare histogram
      } else {
	printf("GetNDCHist(%d) returned NULL\n", ch);
      }
      Fill(&fNDCGroupList, ch, time, 0, 0);//filling histogram group
    }
    //
    //    Int_t pair_ch=GetChannelPair(ch);
    //    if(IsHistDefined(ch)&&IsHistDefined(pair_ch)){ //x=ch, y=Pair(ch)
    //      Fill(&fScatHistGroupList, ch, 0, pair_ch, 0);
    //    }
};

//Decode multi events in 1 trigger pulse
void Run44Decoder::decode_MLFdata(const UChar_t * const p, Int_t nbytes) {
    Int_t nhit = 0;
    MLFdatum * pp = (MLFdatum *)p;
    MLFdatum last_niki_tdc_stop;
    Int_t nitem = nbytes / 8;
    //    printf("at decode_MLFdata: fNch %d\n",fNch);
    for (int i=0; i<nitem; i++) {
	switch (pp[i].GetType()) {
	case 0x11:
	  SetRG(pp[i].GetTime());
	  break;
	case 0x22:
            Push(pp[i],-1);
	    nhit++;
	    break;
	case 0x5b:
	    fT0 = pp[i].GetVal();
	    break;
	case 0x5c:
	    fClock = pp[i].GetVal();
	    break;
	case 0x40: /* niki word */
	    /* nothing to do */
	    break;
	case 0x42: /* niki TDC */
	  last_niki_tdc_stop = pp[i];
	  //	  Push(last_niki_tdc_stop, -1);
	  break;
	case 0x44: /* niki ADC */
	  //	  last_niki_tdc_stop = pp[i];
	  Push(last_niki_tdc_stop, pp[i].GetADC());
	  nhit++;
	  break;
	case 0x46: /* niki TDCADC */
	  last_niki_tdc_stop = pp[i];
	  Push(last_niki_tdc_stop, pp[i].GetADC());
	  nhit++;
	  break;
	}
    }
   for(Int_t ch=0; ch<fNch; ch++){
     //   Int_t ch=pp[i].Get
      Int_t pair_ch=GetChannelPair(ch);
      if(IsHistDefined(ch)&&IsHistDefined(pair_ch)){ //x=ch, y=Pair(ch)
	Fill(&fScatHistGroupList, ch, 0, pair_ch, 0);
	//      fAnyHitList[ch].Print();fAnyHitList[pair_ch].Print();
      }
    }
    if (fNevent < 100)
	printf(" %d bytes contains %d hits\n", nbytes, nhit);
};

