#ifndef _Run44Decoder_h_
#define _Run44Decoder_h_
#include <assert.h>

#include "Hit.h"
#include "CoincidencedHit.h"
#include "Coincidence2.h"

#include <TObjArray.h>
#include <TString.h>

#include <TH1I.h>
#include <TH1D.h>
#include <TH2I.h>
#include <TClonesArray.h>

#include <TSystem.h>
#include <TDirectory.h>

#include "ChGroup.h"
#include "ChGroup2D.h"
#include "Asymmetry.h"

#include "module_numbers.h" 

class MLFdatum;

const Int_t default_nbins_t = 64000 / 100;
const Int_t default_nbins_a = 8192; 

class Run44Decoder : public TNamed {
private:
    const Int_t fNch;
    Int_t fRG;
    TObjArray fSingleHits; // Single Hit list
    TObjArray fDoubleHits; // Double Hit list
    TObjArray fAnyHits; // Single Hits list integrated in all channels
    TObjArray * fSingleHitList; // Hitlist for each channels
    TObjArray * fDoubleHitList; // Hitlist for each channels
    TObjArray * fAnyHitList; // Hitlist for each channels
    Int_t * fPair; // Channel pair for Coincidence and 2D mapping
    TH2I fH2Any; // HitMap of single hits of any kind
    TH2I fH2single; // HitMap of single hits
    TH2I fH2double; // HitMap of double hits
    TH2I fH2TDCADC; // HitMap of TDC vs ADC
    TH2I fH2TDC; // HitMap of TDC hits
    TH2I fH2ADC; // HitMap of ADC hits
    TH2I fH2NDC; // HitMap of NDC hits
    TClonesArray fSingle; // Histograms for each channel
    TClonesArray fDouble; // Histograms for each channel
    TClonesArray fTDC; // Histograms for each channel
    TClonesArray fADC; // Histograms for each channel
    TClonesArray fNDC; // Histograms for each channel
    TObjArray fSingleGroupList;
    TObjArray fDoubleGroupList;
    TObjArray fTDCCutGroupList;
    TObjArray fADCCutGroupList;
    TObjArray fTDCGroupList;
    TObjArray fADCGroupList;
    TObjArray fNDCGroupList;
    TObjArray fScatHistGroupList;
    TObjArray fSingleAsymmetryList;
    TObjArray fDoubleAsymmetryList;
    TString * fCounterMap;
    TString * fCounterMapFilename;
    Bool_t fUseNewCoincidenceAlgorithm;
public:
    Int_t fNevent;
    Int_t fNeventWithBeam;
    Int_t fNhit_single_per_100;
    Int_t fNhit_double_per_100;
    Int_t fNhit_positron_per_100;
    Int_t fNhit_muon_per_100;
    Int_t fNhit_singleFWD_per_100;
    Int_t fNhit_singleBWD_per_100;
    Int_t fNhit_singleLFT_per_100;
    Int_t fNhit_singleRIT_per_100;
    Int_t fNhit_singleUP_per_100;
    Int_t fNhit_singleDWN_per_100;
    Int_t fNhit_doubleFWD_per_100;
    Int_t fNhit_doubleBWD_per_100;
    Int_t fNhit_doubleLFT_per_100;
    Int_t fNhit_doubleRIT_per_100;
    Int_t fNhit_doubleUP_per_100;
    Int_t fNhit_doubleDWN_per_100;
    UInt_t fNhit_single;
    UInt_t fNhit_double;
    ULong64_t fT0; // from GateNet
    ULong64_t fClock; // from GateNet
    char CounterMap[TOTAL_CHANNELS][80];
    char CounterMapFilename[TOTAL_CHANNELS][80];
    double      zero[TOTAL_CHANNELS];
    double      gain[TOTAL_CHANNELS];
    int       CutCh[TOTAL_CHANNELS][MAX_RULES];
    char   CutSubCh[TOTAL_CHANNELS][MAX_RULES];
    char      CutOn[TOTAL_CHANNELS][MAX_RULES];
    double CutRange1[TOTAL_CHANNELS][MAX_RULES];
    double CutRange2[TOTAL_CHANNELS][MAX_RULES];
    ChGroup *CalHist[TOTAL_CHANNELS];
    ChGroup *CutHist[TOTAL_CHANNELS][MAX_RULES];
    ChGroup *ScatHistx[TOTAL_CHANNELS][MAX_RULES];
    ChGroup *ScatHisty[TOTAL_CHANNELS][MAX_RULES];
    ChGroup2D *ScatHist[TOTAL_CHANNELS][MAX_RULES];
    int  MaxRule[TOTAL_CHANNELS];
    int  MaxRule2D[TOTAL_CHANNELS];
    int  IsCal[TOTAL_CHANNELS];


    Run44Decoder(Int_t nch = TOTAL_CHANNELS, Int_t nbins_t = default_nbins_t, Int_t nbins_a = default_nbins_a) :
	fNch(nch),
	fRG(0),
	fSingleHitList(new TObjArray[fNch]),
	fDoubleHitList(new TObjArray[fNch]),
	fAnyHitList(new TObjArray[fNch]),
	fPair(new Int_t[fNch]),
	fH2Any("h2Any", "2D Histogram of single hit of any kinds", nbins_t, 0, 64000, nch, 0, nch),
	fH2single("h2single", "2D Histogram of single hit", nbins_t, 0, 64000, nch, 0, nch),
	fH2double("h2double", "2D Histogram of double hit", nbins_t, 0, 64000, nch, 0, nch),
	fH2TDCADC("h2VME_TDC_vs_ADC", "2D Histogram of VME TDC vs ADC hits", nbins_t, 0, 64000, nbins_a, 0, nbins_a),
#define NIKI_MxC (N_NIKI_3N00_MODULES*N_NIKI_3N00_CHANNELS/2)
	fH2TDC("h2VME_TDC", "2D Histogram of VME TDC hit", nbins_t, 0, 64000, NIKI_MxC, 0, NIKI_MxC),
	fH2ADC("h2VME_ADC", "2D Histogram of VME ADC hit", nbins_a, 0, nbins_a, NIKI_MxC, 0, NIKI_MxC),
#define NIM_MxC (N_NIM_TDC_MODULES*N_NIM_TDC_CHANNELS)
	fH2NDC("h2NIM_TDC", "2D Histogram of NIM TDC hit", nbins_t, 0, 64000, NIM_MxC, 0, NIM_MxC),
	  //	fH2TDC("h2Niki_TDC", "2D Histogram of Niki TDC hit", nbins_t, 0, 64000, nch, 0, nch),
	  //	fH2ADC("h2Niki_ADC", "2D Histogram of Niki ADC hit", nbins_t, 0, 64000, nch, 0, nch),
	  //	fH2NDC("h2Nim_TDC", "2D Histogram of NIM TDC hit", nbins_t, 0, 64000, nch, 0, nch),
	fSingle(TH1D::Class(), fNch*MAX_RG, kTRUE),
	fDouble(TH1D::Class(), fNch*MAX_RG, kTRUE),
	fTDC(TH1D::Class(), fNch*MAX_RG, kTRUE),
	fADC(TH1D::Class(), fNch*MAX_RG, kTRUE),
	fNDC(TH1D::Class(), fNch*MAX_RG, kTRUE),
	fNevent(0),
	fNeventWithBeam(0),
	fNhit_single_per_100(0),
	fNhit_double_per_100(0),
	fNhit_positron_per_100(0),
	fNhit_muon_per_100(0),
	fNhit_single(0),
	fNhit_double(0),
	fNhit_singleFWD_per_100(0),
	fNhit_singleBWD_per_100(0),
	fNhit_singleLFT_per_100(0),
	fNhit_singleRIT_per_100(0),
	fNhit_singleUP_per_100(0),
	fNhit_singleDWN_per_100(0),
	fNhit_doubleFWD_per_100(0),
	fNhit_doubleBWD_per_100(0),
	fNhit_doubleLFT_per_100(0),
	fNhit_doubleRIT_per_100(0),
	fNhit_doubleUP_per_100(0),
	fNhit_doubleDWN_per_100(0),
	  fCounterMap(new TString [fNch]),
	  fCounterMapFilename(new TString [fNch]),
	fUseNewCoincidenceAlgorithm(kTRUE)
    {
      printf("Entering Run44Decoder.h\n");
	gDirectory->cd("/");
	SetNameTitle("Decoder", "Run44Decoder");
	gDirectory->Add(this);

	// memset(fPair, -1, sizeof(fPair));
	for (int i=0; i<fNch; i++) {
	    fPair[i] = -1;
	    strcpy(CounterMap[i],"");
	    IsCal[i]=0;
	}
	/* Channal Pair configuration */
	int i=0;

	/* COPPER Lite */
	if(N_COPPER_LITE_MODULES) 
	  i+=assignPairMap(0,N_COPPER_LITE_CHANNELS,N_COPPER_LITE_MODULES,"/home/daq/DaqComponents/src/Run44/COPPER_LITE.map");
	/* Kalliope */
	if(N_KALLIOPE_TDC_MODULES) 
	  i+=assignPairMap(i,N_KALLIOPE_TDC_CHANNELS,N_KALLIOPE_TDC_MODULES,"KALLIOPE_TDC.map");
	/* NIM_TDC */
	if(N_NIM_TDC_MODULES) 
	  i+=assignHistMap(i,N_NIM_TDC_CHANNELS,N_NIM_TDC_MODULES,"NIM_TDC.map");

	/* Niki VME */
	if(N_NIKI_3N00_MODULES) 
	  i+=assignHistMap(i,N_NIKI_3N00_CHANNELS,N_NIKI_3N00_MODULES,"NIKI_3N00.map");

	fTDCCutGroupList.SetOwner(kTRUE);
	fADCCutGroupList.SetOwner(kTRUE);
	fTDCGroupList.SetOwner(kTRUE);
	fADCGroupList.SetOwner(kTRUE);
	fNDCGroupList.SetOwner(kTRUE);
	fScatHistGroupList.SetOwner(kTRUE);
	fSingleHits.SetOwner(kTRUE);
	fAnyHits.SetOwner(kTRUE);

	for(int p=0; p<MAX_RG; p++) {
	  for (int i=0; i<fNch; i++) {

	  fSingleHitList[i].SetOwner(kTRUE);
	  fDoubleHitList[i].SetOwner(kTRUE);
	  fAnyHitList[i].SetOwner(kTRUE);
	  
	  TString name, title;
	  TString temp;
	  int BaseCh;
	  temp=GetCounterMapFilename(i);
	  //	    cerr<<i<<": "<<temp.Data()<<endl;
	  
	  BaseCh=0;
	  if(IsKalliope(i)||IsCopperLite(i)){
	    //	      printf("single/double\n");
	    name.Form("single_ch%04d_%1d", i-BaseCh,p);
	    title.Form("single hit histogram of ch%04d on %1d-cond.", i,p);
	    new (fSingle.AddrAt(i+fNch*p)) TH1D(name, title, nbins_t, 0, 64000);
	    
	    name.Form("double_ch%04d_%1d", i-BaseCh,p);
	    title.Form("double hit histogram of ch%04d on %1d-cond.", i,p);
	    new (fDouble.AddrAt(i+fNch*p)) TH1D(name, title, nbins_t, 0, 64000);
	  }
	  
	  BaseCh=N_COPPER_LITE_MODULES*N_COPPER_LITE_CHANNELS
	    +N_KALLIOPE_TDC_MODULES*N_KALLIOPE_TDC_CHANNELS;
	  if(IsNim(i)){
	    //	      	      printf("NDC\n");
	    name.Form("NIM_TDC_%03dN_%1d", i-BaseCh,p);
	    title.Form("NIM TDC histogram of ch%03dN on %1d-cond.", i-BaseCh,p);
	    new (fNDC.AddrAt(i+fNch*p)) TH1D(name, title, nbins_t, 0, 64000);
	    if(IsCal[i]){
	      GetNDCGroupList()->Add(CalHist[i]);
	      for(int j=0; j<MaxRule[i]; j++) GetNDCGroupList()->Add(CutHist[i][j]);
	    }
	  }
	  
	  BaseCh=N_COPPER_LITE_MODULES*N_COPPER_LITE_CHANNELS
	    +N_KALLIOPE_TDC_MODULES*N_KALLIOPE_TDC_CHANNELS
	    +N_NIM_TDC_MODULES*N_NIM_TDC_CHANNELS;
	  if(IsNiki(i)){
	    //	      printf("ADC/TDC\n");
	    if((i-BaseCh)%2){
	      name.Form("VME_ADC_%03dA_%1d", (i-BaseCh)/2,p);
	      title.Form("Niki ADC histogram of ch%03dA on %1d-cond.", (i-BaseCh)/2,p);
	      new (fADC.AddrAt(i+fNch*p)) TH1D(name, title, nbins_a, 0, nbins_a);
	      if(IsCal[i]){
		GetADCGroupList()->Add(CalHist[i]);
		for(int j=0; j<MaxRule[i]; j++) GetADCGroupList()->Add(CutHist[i][j]);
	      }
	    }
	    else{
	      name.Form("VME_TDC_%03dT_%1d", (i-BaseCh)/2,p);
	      title.Form("Niki TimeStamp histogram of ch%03dT on %1d-cond.", (i-BaseCh)/2,p);
	      new (fTDC.AddrAt(i+fNch*p)) TH1D(name, title, nbins_t, 0, 64000);
	      if(IsCal[i]){
		GetTDCGroupList()->Add(CalHist[i]);
		for(int j=0; j<MaxRule[i]; j++) GetTDCCutGroupList()->Add(CutHist[i][j]);
	      }
	    }
	  }
	  if(IsCal[i]) 
	    for(int j=0; j<MaxRule2D[i]; j++)
	      if(IsCal[fPair[i]]) GetScatHistGroupList()->Add(ScatHist[i][j]);
	  }
	}
	printf("at Run44Decoder: %d\n",fNch);
	printf("Exitting Run44Decoder.h\n");
    };
	
ChGroup *
newgroup(const char * name, const char * title)
{
  ChGroup * g = (ChGroup*)gDirectory->FindObject(name);
    if (g == NULL) {
      printf("Total_channels: %d\n",fNch);
	g = new ChGroup(name, title, fNch);
	gDirectory->Add(g);
    }else{
      g->SetTitle(title);
      gDirectory->Add(g);
    }
    return g;
}
ChGroup2D *
newgroup2D(const char * name, const char * title)
{
  ChGroup2D * g = (ChGroup2D*)gDirectory->FindObject(name);
    if (g == NULL) {
	g = new ChGroup2D(name, title, fNch);
	gDirectory->Add(g);
    }else{
      g->SetTitle(title);
      gDirectory->Add(g);
    }
    return g;
}

  int assignPairMap(int BaseCh, int N_CHANNELS, int N_MODULES, char *MapFilename)
  {
    char line[256];
    FILE *fpMap;
    int NoCounters,PairIncr;
    fpMap=fopen(MapFilename,"r");
    if(fpMap==NULL){
      printf("PairMapFile: %s does not exist.\n",MapFilename);
      exit(1);
    }
    else{
      printf("\nReading %s ...\n",MapFilename);
      fgets(line,255,fpMap);
      printf("Spectrometer: %s\n",line);
      fgets(line,255,fpMap); sscanf(line,"%d-%d",&NoCounters,&PairIncr);
      printf("Number of Counters: %d\n",NoCounters);
      printf("Pair Increment Step: %d\n",PairIncr);
    }
    if(N_CHANNELS*N_MODULES!=NoCounters)
      {
	printf("Inconsistent No. of Counters: %d vs. %d.\n",N_CHANNELS*N_MODULES,NoCounters);
	//	exit(1);
      }
    int i,ii,iii,t0dummy;
    int ip=1,ip_save=-1;
    char MapList[80],range[80];

    for(int j=0;j<NoCounters/2;j++)
      if(fgets(line,255,fpMap)){
	if(strpbrk(line,":")){
	  sscanf(line,"%d:%s",&ip,line);
	  if(ip!=ip_save) printf("#%d board---------------------------------------------\n",ip);
	  ip_save=ip;
	}
	if(strpbrk(line,"-")){
	  sscanf(line,"%d-%d,%d,%s",&i,&ii,&t0dummy,MapList);
	  i+=N_CHANNELS*(ip-1)+BaseCh;ii+=N_CHANNELS*(ip-1)+BaseCh;

	  if(abs(ii-i)!=PairIncr){
	    printf("Inconsistent Pair %d-%d with Increment %d\n",i,ii,PairIncr);
	  }
	  iii=GetChannelPair(i);
	  if (iii==-1) {
	    SetChannelPair(i, ii);strcpy(CounterMap[i],MapList);strcpy(CounterMap[ii],MapList);
	    strcpy(CounterMapFilename[i],MapFilename);strcpy(CounterMapFilename[ii],MapFilename);
	    //cerr << "New Pair:" << i << ',' << ii << ',' << CounterMap[i] << endl;
	  } else {
	    printf("Redifinition of Pair for %d: \n",i);
	    exit(1);
	  }
	}
	else{
	  sscanf(strpbrk(line,",")+1,"%d,%s",&t0dummy,MapList);
	  line[strpbrk(line,",")-line]='\0';
	  if(!strcmp(line,"all"))
	    for(int k=0;k<N_CHANNELS;k++){
	      i=k+N_CHANNELS*(ip-1)+BaseCh;ii=i+PairIncr;
	      iii=GetChannelPair(i);
	      if (iii==-1) {
		SetChannelPair(i, ii);strcpy(CounterMap[i],MapList);strcpy(CounterMap[ii],MapList);
		strcpy(CounterMapFilename[i],MapFilename);strcpy(CounterMapFilename[ii],MapFilename);
		//cerr << "Gen Pair:" << i << ',' << ii << ',' << CounterMap[i] << endl;
	      }
	    }
	}
      }
      else{
	break;
      }
    for(int k=0;k<NoCounters;k++){
      if(GetChannelPair(k+BaseCh)==-1) {
	printf("Insufficient Counter Map: %s for channel %d\n",MapFilename,k+BaseCh);
	exit(1);
      }
      else{
	SetCounterMap(BaseCh+k);
	SetCounterMapFilename(BaseCh+k);
	TString temp;
	//	temp=GetCounterMapFilename(BaseCh+k);
     	//	printf("%s:",temp.Data());
	//	temp=GetCounterMap(BaseCh+k);
	//	printf("%s:%s\n",CounterMap[BaseCh+k],temp.Data());
	
      }
    }
    fclose(fpMap);
    return NoCounters;
  };

  int assignHistMap(int BaseCh, int N_CHANNELS, int N_MODULES, char *MapFilename)
  {
    int N_SUBCH=1;
        if(!strcmp(MapFilename,"NIKI_3N00.map"))N_SUBCH=2;

    char line[256];
    FILE *fpMap;
    int NoCounters,PairIncr;
    fpMap=fopen(MapFilename,"r");
    if(fpMap==NULL){
      printf("HistMapFile: %s does not exist.\n",MapFilename);
      exit(1);
    }
    else{
      printf("\nReading %s ...\n",MapFilename);
      fgets(line,255,fpMap);
      printf("Spectrometer: %s\n",line);
      fgets(line,255,fpMap); sscanf(line,"%d*%d",&NoCounters,&PairIncr);
      printf("Number of Counters: %d\n",NoCounters);
      printf("Number of Sub-channels per Counter: %d\n",PairIncr);
    }
    if(N_CHANNELS*N_MODULES!=NoCounters){
      printf("Inconsistent No. of Counters: %d vs. %d.\n",N_CHANNELS*N_MODULES,NoCounters);
      exit(1);
    }
    if(N_SUBCH!=PairIncr){
      printf("Inconsistent No. of Sub channels: %d vs. %d.\n",N_SUBCH,PairIncr);
      exit(1);
    }
    int i,ii,iii,t0dummy;
    int ip=1,ip_save=-1;
    char * pos;
    double a,b;
    char c,h,hh,hhh,MapList[80],range[80];
    TString name, title;
    TString temp;

    //    int k=0;
    //    int l=0;

    int n=-1;
    int m=-1; 
    int p=-1;
    int k=0;
    while(fgets(line,255,fpMap)){
      pos=strpbrk(line,":");
      if(pos){
	n++;
	*pos='\0';  sscanf(line,"%d%c",&i,&c);
	//	  printf("%d%c\n",i,c);
	if(N_SUBCH==2){
	  if( n!=i*2 && n!=2*i+1){
	    printf("Invalid channel number in %s: %d%c for %d.\n",MapFilename,i,c,n);
	    exit(1);
	  }
	  if((n%2&&c=='T')||(!(n%2)&&c=='A')){
	    printf("Invalid order of channels in %s: %d%c for %d.\n",MapFilename,i,c,n);
	    exit(1);
	  }	 
	} 
	k=n+BaseCh;
	IsCal[k]=0;
	CalHist[k]=NULL;
	for(int mm=0; mm<MAX_RULES;mm++) CutHist[k][mm]=NULL;
	switch(c){
	case 'T': temp="TDC_"; strcat(CounterMap[k],"T"); break;
	case 'A': temp="ADC_"; strcat(CounterMap[k],"A"); break;
	case 'N': temp="NDC_"; strcat(CounterMap[k],"N"); break;
	default :
	  printf("Unknown channel attribute: %c at line %d%c:%s\n",c,i,c,line);
	  exit(1);
	}
	
	strcpy(line,pos+1);
	h=NULL;
	pos=strpbrk(line,">");
	if(pos){
	  IsCal[k]++;
	  *pos='\0'; //*(pos+1)='\0'; 
	  sscanf(line,"%lf,%lf",&a,&b);
	  strcpy(line,pos+2);
	  int j=0;
	  while(h==NULL){
	    if(line[j]>='A'&&line[j]<='Z') h=line[j];
	    j++;
	  }
	  name=temp+h;
	  ChGroup * g = (ChGroup*)gDirectory->FindObject(name);
	  if(g!=NULL){
	    title=g->GetTitle();
	  }else{
	    title="Histograms: ";
	  }
	  char t[80];
	  sprintf(t,"(%d%c-%5.2f)x%5.2f ",i,*name.Data(),a,b);
	  title+=t;
	  //	    cerr<< title.Data() <<endl;
	  //	    n++;
	  CalHist[k]=newgroup(name.Data(),title.Data());
	  //	    CalHist[k]->Print();
	  CalHist[k]->GetInfo()->AddCh(k);
	  CalHist[k]->GetInfo()->SetZeroGain(k,a,b);
	  zero[k]=a;gain[k]=b;
	}
	else{
	  //printf("k:%d %s\n",k,CounterMap[k]);
	  //	  strcpy(CounterMapFilename[k],MapFilename);
	  //	  SetCounterMap(k);
	  //	  SetCounterMapFilename(k);
	  zero[k]=0;gain[k]=1;
	}
	m=-1; p=-1;
      }
      else{
	//	  printf(line);
	pos=strpbrk(line,"/");
	if(pos){
	  m++;
	  if(m+1>MAX_RULES){
	    printf("Exceeded Maximum numbers of Rules %d > %d\n",m+1,MAX_RULES);
	    exit(1);
	  }
	  strcpy(line,pos+1);
	  pos=strpbrk(line,">");
	  if(pos){
	    *pos='\0'; //*(pos+1)='\0'; 
	    sscanf(line,"%d%c=%c(%lf,%lf)",&ii,&c,&h,&a,&b);
	    strcpy(line,pos+2);
	    hh=NULL;
	    int j=0;
	    while(hh==NULL){
	      if(line[j]>='A'&&line[j]<='Z') hh=line[j];
	      j++;
	    }
	    name=temp+hh;
	    ChGroup  * g = (ChGroup*) gDirectory->FindObject(name);
	    if(g!=NULL){
	      title=g->GetTitle();
	    }else{
	      //	    cerr<<name.Data()<<endl;
	      title="Histograms: ";
	    }
	    char t[80];
	    sprintf(t,"(%d%c-%5.2lf)x%5.2lf cut on %d%c=%c(%5.2lf,%5.2lf) ",
		    i,*name.Data(),zero[k],gain[k],ii,c,h,a,b);
	    title+=t;
	    CutHist[k][m]=newgroup(name.Data(),title.Data());
	    //	    CutHist[k][m]->Print();
	    CutHist[k][m]->GetInfo()->AddCh(k);
	    CutHist[k][m]->GetInfo()->SetZeroGain(k,zero[k],gain[k]);
	    CutCh[k][m]=ii;CutSubCh[k][m]=c;CutOn[k][m]=h;
	    CutRange1[k][m]=a;CutRange2[k][m]=b;
	    CutHist[k][m]->GetInfo()->SetCutRange(k,a,b);
	    //	    printf("%d%c/%d%c=%c->%c\n",i,*name.Data(),ii,c,h,hh);
	  }
	}
	pos=strpbrk(line,"x");
	if(pos){
	  p++;
	  if(m+1>MAX_RULES){
	    printf("Exceeded Maximum numbers of Rules for 2d %d > %d\n",p+1,MAX_RULES);
	    exit(1);
	  }
	  strcpy(line,pos+1);
	  char t[80];
	  hhh=NULL; strcpy(t,"");
	  int j=-1;
	  while(hhh==NULL||hhh==' '){
	    j++;
	    hhh=line[j];
	  }
	  strcpy(t,line+j);t[strlen(t)-1]=NULL;
	  //	  name=CalHist[k]->GetName();
	  name=t;
	  ChGroup  * g = (ChGroup*) gDirectory->FindObject(name);
	  if(g==NULL){
	    cerr<<"Could not find y-axis of 2D Histogram "<<name.Data()
		<<" for x-axis "<<CalHist[k]->GetName()<<endl;
	    exit(1);
	  }else{
	    title=g->GetTitle();
	  }
	  ScatHistx[k][p]=CalHist[k];
	  ScatHisty[k][p]=g;
	  name=ScatHistx[k][p]->GetName();name+="_x_";name+=ScatHisty[k][p]->GetName();
	  title="2D Histogram of ";
	  title+=ScatHistx[k][p]->GetName();title+=" vs ";
	  title+=ScatHisty[k][p]->GetName();
	  ScatHist[k][p]=newgroup2D(name.Data(),title.Data());
	  ScatHist[k][p]->SetInfox(ScatHistx[k][p]->GetInfo());
	  ScatHist[k][p]->SetInfoy(ScatHisty[k][p]->GetInfo());
	}
      }
      MaxRule[k]=m+1;
      MaxRule2D[k]=p+1;
    }
    fclose(fpMap);

    if(n+1!=NoCounters){
      printf("Inconsistent lines: %d vs. %d.\n",n+1,NoCounters);
      exit(1);
    }
    for(int n=0;n<NoCounters;n++){
      k=n+BaseCh;
      strcpy(CounterMapFilename[k],MapFilename);
      SetCounterMap(k);
      SetCounterMapFilename(k);

      //      cerr<<k<<":"<<IsCal[k]<<endl;

      if(IsCal[k]){ 
	CalHist[k]->Print(); 
	for(int m=0;m<MaxRule[k];m++){
	  CutHist[k][m]->Print();
	  //      	GetNDCGroupList()->Add(CutHist[k][m]);	
	}
	for(int p=0;p<MaxRule2D[k];p++){
	  ScatHist[k][p]->Print();
	  for(int q=0; q<TOTAL_CHANNELS; q++){
	    if(IsCal[q]){
	      //	      ScatHisty[k][p]->Print();
	      if(strcmp(ScatHisty[k][p]->GetName(),CalHist[q]->GetName())==0){
		fPair[k]=q;
	      }
	    }
	  //      	GetScatHistGroupList()->Add(ScatHist[k][p]);	
	  //	ScatHistx[k][p]->GetInfo()->Print();
	  //	ScatHisty[k][p]->GetInfo()->Print();
	  }
	}
      }
    }
    cerr<<"exitting assignHistMap"<<endl;
    return NoCounters;
  };

    void Clear(Option_t * option = "") {
	fSingleHits.Clear();
	fDoubleHits.Clear();
	fAnyHits.Clear();
	for (int i=0; i<fNch; i++) {
	  fSingleHitList[i].Clear(option);
	  fDoubleHitList[i].Clear(option);
	  fAnyHitList[i].Clear(option);
	}
    };

    ~Run44Decoder() {
	Clear();
	delete [] fSingleHitList;
	delete [] fDoubleHitList;
	delete [] fAnyHitList;
	delete [] fPair;
	fSingleHitList = NULL;
	fDoubleHitList = NULL;
	fAnyHitList = NULL;
	fPair = NULL;
    };

    Bool_t IsCopperLite(Int_t ch) {
      if(strcmp(fCounterMapFilename[ch],"COPPER_LITE.map")){
	return kFALSE;
      }
      else{
	return kTRUE;
      }
    };
    Bool_t IsNiki(Int_t ch) {
      if(strcmp(fCounterMapFilename[ch],"NIKI_3N00.map")){
	return kFALSE;
      }
      else{
	return kTRUE;
      }
    };
    Bool_t IsKalliope(Int_t ch) {
      if(strcmp(fCounterMapFilename[ch],"KALLIOPE_TDC.map")){
	return kFALSE;
      }
      else{
	return kTRUE;
      }
    };
    Bool_t IsNim(Int_t ch) {
      if(strcmp(fCounterMapFilename[ch],"NIM_TDC.map")){
	return kFALSE;
      }
      else{
	return kTRUE;
      }
    };
    Bool_t IsHistDefined(Int_t ch) {
      if(IsCal[ch]==0){
	return kFALSE;
      }
      else{
	return kTRUE;
      }
    };

    TString  GetCounterMap(int i){
      return fCounterMap[i];
    };
    void SetCounterMap(int i){
      fCounterMap[i]=CounterMap[i];
    }
    TString  GetCounterMapFilename(int i){
      return fCounterMapFilename[i];
    };
    void SetCounterMapFilename(int i){
      fCounterMapFilename[i]=CounterMapFilename[i];
    }

    TCollection * GetSingleGroupList() {
	return &fSingleGroupList;
    };

    TCollection * GetDoubleGroupList() {
	return &fDoubleGroupList;
    };

    TCollection * GetTDCCutGroupList() {
	return &fTDCCutGroupList;
    };

    TCollection * GetADCCutGroupList() {
	return &fADCCutGroupList;
    };

    TCollection * GetTDCGroupList() {
	return &fTDCGroupList;
    };

    TCollection * GetADCGroupList() {
	return &fADCGroupList;
    };

    TCollection * GetNDCGroupList() {
	return &fNDCGroupList;
    };

    TCollection * GetScatHistGroupList() {
	return &fScatHistGroupList;
    };

    TCollection * GetSingleAsymmetryList() {
	return &fSingleAsymmetryList;
    };

    TCollection * GetDoubleAsymmetryList() {
	return &fDoubleAsymmetryList;
    };

    TH1 * GetSingleHist(Int_t ch) {
	return (TH1 *)fSingle.At(ch);
    };

    TH1 * GetDoubleHist(Int_t ch) {
	return (TH1 *)fDouble.At(ch);
    };

    TH1 * GetTDCHist(Int_t ch) {
	return (TH1 *)fTDC.At(ch);
    };

    TH1 * GetADCHist(Int_t ch) {
	return (TH1 *)fADC.At(ch);
    };

    TH1 * GetNDCHist(Int_t ch) {
	return (TH1 *)fNDC.At(ch);
    };

    void Reset(TCollection * col) {
	TIter iter(col);
	while (TObject * o = iter()) {
	    if (o->IsA() == Asymmetry::Class()) {
	      Asymmetry * a = (Asymmetry *)o;
	      a->Reset();//printf("passed a->reset()\n");
	    } else if (o->IsA() == ChGroup::Class()) {
	      ChGroup * g = (ChGroup *)o;
	      g->Reset();//printf("passed g->reset()\n");
	    } else if (o->IsA() == ChGroup2D::Class()) {
	      ChGroup2D * m = (ChGroup2D *)o;
	      m->Reset();//printf("passed m->reset()\n");
	    } else {
	      assert("unknown class");
	    }
	}
    };

void Fill(TCollection * col, Int_t ch, Int_t time, Int_t ch2, Int_t time2) {
	TIter iter(col);
	while (TObject * o = iter()) {
	    if (o->IsA() == Asymmetry::Class()) {
		Asymmetry * a = (Asymmetry *)o;
		a->Fill(ch, time);
	    } else if (o->IsA() == ChGroup::Class()) {
		ChGroup * g = (ChGroup *)o;
		if(ch2==0&&time2==0){
		  g->Fill(ch, time);
		  //		  printf("%d_%d: %d",ch,GetRG(),time); g->Print();
		}else if((double)time2/1000.0 > g->GetInfo()->GetLower(ch) && 
			 (double)time2/1000.0 < g->GetInfo()->GetUpper(ch)){
		  g->Fill(ch, time);
		  //		  printf("%d_%d: %d @ %lf=(%lf,%lf)",ch,GetRG(),time,(double)time2/1000.0,g->GetInfo()->GetLower(ch),g->GetInfo()->GetUpper(ch)); g->Print();
		}
	    } else if (o->IsA() == ChGroup2D::Class()) {
	      ChGroup2D * m = (ChGroup2D *)o;
	      TObjArray * p = GetAnyHitList(ch);
	      TObjArray * q = GetAnyHitList(ch2);
	      for(int i=0; i<p->GetEntries();i++){
		Hit * pp=(Hit *)p->At(i);
		Int_t time1=pp->GetTime();
		for(int j=0; j<q->GetEntries();j++){
		  Hit * qq=(Hit *)q->At(j);
		  Int_t time3=qq->GetTime();
		  //		  cerr<<ch<<" "<<time1<<" "<<ch2<<" "<<time3<<endl;
		  m->Fill2D(ch,time1,ch2,time3);
		}
	      }
	    } else {
		assert("unknown class");
	    }
	}
    };

    void Push(MLFdatum & item, int adc = -1);

    void SetRG(Int_t RG) {
      fRG=RG;
      //      fprintf(stderr,"\n%d: ",RG);
    };

    Int_t GetRG() {
      return fRG;
    };
  

    void decode_MLFdata(const UChar_t * const p, Int_t nbytes) ;

    void Reset() {
	Reset(&fSingleGroupList);
	Reset(&fDoubleGroupList);
	Reset(&fTDCCutGroupList);
	Reset(&fADCCutGroupList);
	Reset(&fTDCGroupList);
	Reset(&fADCGroupList);
	Reset(&fNDCGroupList);
	Reset(&fScatHistGroupList);
	Reset(&fSingleAsymmetryList);
	Reset(&fDoubleAsymmetryList);
	fH2Any.Reset();
	fH2single.Reset();
	fH2double.Reset();
	fH2TDCADC.Reset();
	fH2TDC.Reset();
	fH2ADC.Reset();
	fH2NDC.Reset();
	printf("Passed Reset()\n");
	printf("fNch: %d\n",fNch);
	for (int p=0; p<MAX_RG; p++) {
	for (int i=0; i<fNch; i++) {
	  printf("%d",i);
	  if(IsCopperLite(i)||IsKalliope(i)){
	    GetSingleHist(i+fNch*p)->Reset();printf(".");
	    GetDoubleHist(i+fNch*p)->Reset();printf(".");
	  }
	  if(IsNiki(i)){
	    if(i%2){
	      GetADCHist(i+fNch*p)->Reset();printf(".");
	    }
	    else{
	      GetTDCHist(i+fNch*p)->Reset();printf(".");
	    }
	  }
	  if(IsNim(i)){
	    GetNDCHist(i+fNch*p)->Reset();printf(".\n");
	  }
	}
	}
	printf("Passed ->Reset()\n");

	fNevent = 0;
	fNeventWithBeam = 0;
	fNhit_single_per_100 = 0;
	fNhit_double_per_100 = 0;
	fNhit_positron_per_100 = 0;
	fNhit_muon_per_100 = 0;

	fNhit_singleFWD_per_100 = 0;
	fNhit_singleBWD_per_100 = 0;
	fNhit_singleLFT_per_100 = 0;
	fNhit_singleRIT_per_100 = 0;
	fNhit_singleUP_per_100 = 0;
	fNhit_singleDWN_per_100 = 0;
	fNhit_doubleFWD_per_100 = 0;
	fNhit_doubleBWD_per_100 = 0;
	fNhit_doubleLFT_per_100 = 0;
	fNhit_doubleRIT_per_100 = 0;
	fNhit_doubleUP_per_100 = 0;
	fNhit_doubleDWN_per_100 = 0;

	fNhit_single = 0;
	fNhit_double = 0;
    };

    void SetChannelPair(Int_t ch1, Int_t ch2) {
	fPair[ch1] = ch2;
	fPair[ch2] = ch1;
    };

    Int_t GetChannelPair(Int_t ch1) {
	return fPair[ch1];
    };

    Int_t neighbor_ch_minus(Int_t ch) {

	Int_t ret;
	switch (ch+1) {
	case   1: ret =  33 - 1; break;
	case  33: ret =   1 - 1; break;
	case  65: ret =  97 - 1; break;
	case  97: ret =  65 - 1; break;
	case 129: ret = 161 - 1; break;
	case 161: ret = 129 - 1; break;
	case 193: ret = 225 - 1; break;
	case 225: ret = 193 - 1; break;
	default:
		ret = ch - 1;
	}

	return ret;
    };

    Int_t neighbor_ch_plus(Int_t ch) {

	Int_t ret;
	switch (ch+1) {
	case  32: ret =  64 - 1; break;
	case  64: ret =  32 - 1; break;
	case  96: ret = 128 - 1; break;
	case 128: ret =  96 - 1; break;
	case 160: ret = 192 - 1; break;
	case 192: ret = 160 - 1; break;
	case 224: ret = 256 - 1; break;
	case 256: ret = 224 - 1; break;
	default:
		ret = ch + 1;
	}

	return ret;
    };

    /*    Bool_t is_outer(Int_t ch) {
      int Base,Top;
      Base=0;
      Top=N_COPPER_LITE_CHANNELS*N_COPPER_LITE_MODULES;
      if(ch>=Base && ch<Top){  //ch in COPPER_LITE_MODULES for DOmega-1
	if (ch-Base+1 >= 65 && ch-Base+1 <= 128)
	  return kTRUE;
	if (ch-Base+1 >= 191 && ch-Base+1 <= 256)
	  return kTRUE;
      }
      Base=Top;
      Top=Base+N_NIKI_3N00_CHANNELS*N_NIKI_3N00_MODULES;
      if(ch>=Base && ch<Top){  //ch in NIKI_3N00_MODULES
      }
      Base=Top;
      Top=Base+N_KALLIOPE_TDC_CHANNELS*N_KALLIOPE_TDC_MODULES;
      if(ch>=Base && ch<Top){  //ch in Kalliope_TDC
	if(ch-Base > 0x00 && ((ch-Base) & 0x010))
	  return kTRUE;
      }
      Base=Top;
      Top=Base+N_NIM_TDC_CHANNELS*N_NIM_TDC_MODULES;
      if(ch>=Base && ch<Top){  //ch in NIM_TDC
      }

      return kFALSE;

    };
    */

    void Coincidence(Int_t gap) {

	Bool_t processed[fNch];

	memset(processed, 0, sizeof(processed));

	for (int i=0; i<fNch; i++) {
	  //	    if (is_outer(i))
	  //	      continue;
	    if (processed[i])
	      continue;
	    if (fPair[i] < 0)
	      continue;

	    fDoubleHitList[i].Clear();
	    fDoubleHitList[fPair[i]].Clear();

	    if (fUseNewCoincidenceAlgorithm) {
		CoincidencedHit::FillList2(
		    fSingleHitList[i], fSingleHitList[fPair[i]], gap,
		    fDoubleHitList[i]);
	    } else {
		CoincidencedHit::FillList(
		    fSingleHitList[i], fSingleHitList[fPair[i]], gap,
		    fDoubleHitList[i]);
	    }

	    TIter iter(&fDoubleHitList[i]);
	    while (TObject * o = iter()) {
		assert(o->IsA() == CoincidencedHit::Class());

		fNhit_double_per_100++;
		fNhit_double++;
		TString temp;
		temp=GetCounterMap(i);
		//		printf("%d: %s\n",i,temp.Data());
		if(strpbrk(temp.Data(),"F")) fNhit_doubleFWD_per_100++;
		if(strpbrk(temp.Data(),"B")) fNhit_doubleBWD_per_100++;
		if(strpbrk(temp.Data(),"L")) fNhit_doubleLFT_per_100++;
		if(strpbrk(temp.Data(),"R")) fNhit_doubleRIT_per_100++;
		if(strpbrk(temp.Data(),"U")) fNhit_doubleUP_per_100++;
		if(strpbrk(temp.Data(),"D")) fNhit_doubleDWN_per_100++;

		fDoubleHits.Add(o);

		CoincidencedHit * hit = (CoincidencedHit *)o;
		const Hit & hit1 = hit->GetHit1();
		const Hit & hit2 = hit->GetHit2();
		GetDoubleHist(hit1.GetCh()+fNch*fRG)->Fill(hit1.GetTime());
		GetDoubleHist(hit2.GetCh()+fNch*fRG)->Fill(hit2.GetTime());
		fH2double.Fill(hit1.GetTime(), hit1.GetCh());
		fH2double.Fill(hit2.GetTime(), hit2.GetCh());

		Fill(&fDoubleGroupList, hit1.GetCh(), hit1.GetTime(), 0, 0);
		Fill(&fDoubleGroupList, hit2.GetCh(), hit2.GetTime(), 0, 0);
		Fill(&fDoubleAsymmetryList, hit1.GetCh(), hit1.GetTime(), 0, 0);
		Fill(&fDoubleAsymmetryList, hit2.GetCh(), hit2.GetTime(), 0, 0);
	    }

	    processed[i] = kTRUE;
	    processed[fPair[i]] = kTRUE;
	}
    };

    TObjArray * GetCoincidenced(Int_t ch) {

	if (fPair[ch] == -1)
	    return NULL;

	TObjArray * list = &fDoubleHitList[ch];
	if (list->GetEntries() > 0)
	    return list;
	return &fDoubleHitList[fPair[ch]];
    };

    void Print(Option_t * option = "") const {
	fH2Any.Print(option);
	fH2single.Print(option);
	fH2TDCADC.Print(option);
	fH2TDC.Print(option);
	fH2ADC.Print(option);
	fH2NDC.Print(option);
    };

    void Draw(Option_t * opt = "colz") {
	fH2TDCADC.Draw(opt);
	fH2TDC.Draw(opt);
	fH2ADC.Draw(opt);
	fH2NDC.Draw(opt);
	fH2single.Draw(opt);
	fH2Any.Draw(opt);
	// fH2single.GetYaxis()->SetRangeUser(0,0x2FF);
    };

    TH2 * GetRawHist() {
	return &fH2single;
    };

    TH2 * GetTDCHist() {
	return &fH2TDC;
    };

    TH2 * GetADCHist() {
	return &fH2ADC;
    };

    TH2 * GetNDCHist() {
	return &fH2NDC;
    };

    TH2 * GetCoincidencedHist() {
	return &fH2double;
    };

    TObjArray * GetSingleHits() {
	return & fSingleHits;
    };

    TObjArray * GetDoubleHits() {
	return & fDoubleHits;
    };

    TObjArray * GetAnyHits() {
	return & fAnyHits;
    };

    TObjArray * GetAnyHitList(Int_t ch) {
	return & fAnyHitList[ch];
    };

    Int_t GetNCoindicencedHit() {
	Int_t nhit = 0;
	for (int ch=0; ch<fNch; ch++) {
	    nhit += fDoubleHitList[ch].GetEntries();
	}
	return nhit;
    };

    Int_t GetNch() const {
	return fNch;
    };

    void UpdateAsymmetry() {

	TIter singles(&fSingleAsymmetryList);
	while (TObject * o = singles()) {
	    assert(o->IsA() == Asymmetry::Class());
	    Asymmetry * a = (Asymmetry *)o;
cout << "UpdateAsymmetry, " << a->GetName() << " Update" << endl;
	    a->Update();
	}

	TIter doubles(&fDoubleAsymmetryList);
	while (TObject * o = doubles()) {
	    assert(o->IsA() == Asymmetry::Class());
	    Asymmetry * a = (Asymmetry *)o;
cout << "UpdateAsymmetry, " << a->GetName() << " Update" << endl;
	    a->Update();
	}

    };

    void UseNewCoincidenceAlgorithm(Bool_t flag) {
	fUseNewCoincidenceAlgorithm = flag;
    };

    Bool_t UseNewCoincidenceAlgorithm() const {
	return fUseNewCoincidenceAlgorithm;
    };

    ClassDef(Run44Decoder, 0)


};
#endif

