#ifndef _RPV130_H_
#define _RPV130_H_

#include <unistd.h>

#include <assert.h>
#include <fcntl.h>
#include <sys/mman.h>

#include <Rtypes.h>
#include <Riostream.h>

const int mapsize = 16;

union rpv130_csr {
    UShort_t val;
    struct {
        UInt_t busy3:1;         /* 0x80 */
        UInt_t enable3:1;       /* 0x40 */
        UInt_t busy:1;          /* 0x20 */
        UInt_t enable:1;        /* 0x10 */
        UInt_t pad:2;           /* 0x04, 0x08 */
        UInt_t clear:1;         /* 0x02 */
        UInt_t clear3:1;        /* 0x01 */
    };
};

struct rpv130_register {
    UShort_t    latch1;         /* 0, in */
    UShort_t    latch2;         /* 2, in */
    UShort_t    flipflop;       /* 4, in */
    UShort_t    through;        /* 6, in */
    UShort_t    out;            /* 8, out */
    UShort_t    level;          /* A, out */
    rpv130_csr  csr1;           /* C */
    rpv130_csr  csr2;           /* E */
};

class rpv130 {
private:
    int fd;
    int size;
    volatile rpv130_register *mma;

    caddr_t vme_mapopen(off_t start, size_t size)
    {
        void *addr =
            mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, start);
        assert(addr != MAP_FAILED);
        return (caddr_t)addr;
    };

public:

    rpv130() : fd(-1), size(0), mma(NULL) { };

    rpv130(const int start) :
        fd(open("/dev/vmedrv16d16", O_RDWR)), 
        size(mapsize),
        mma((rpv130_register *)vme_mapopen(start, size))
    {
        assert(fd != -1);
        assert(mma != MAP_FAILED);
    };

    virtual ~rpv130() {
        if (fd != -1) {
            munmap((void *)mma, size);
            close(fd);
        }
    };

    void Open(const int start) {
        assert(fd == -1);
        fd = open("/dev/vmedrv16d16", O_RDWR);
        size = mapsize;
        mma = (rpv130_register *)vme_mapopen(start, size);
    };

    void Close() {
        assert(fd != -1);
        munmap((void *)mma, size);
        close(fd);
        fd = -1;
    };

    virtual void Print(Option_t * opt = "") const {
        assert(fd != -1);
        assert(mma != MAP_FAILED);
        assert(mma != NULL);
        UShort_t * usp = (UShort_t *)mma;
        for (int i=0; i<0x10; i+=2) {
            printf("%02x %04x\n", (int)i*2, (int)usp[i]);
        }
    };

    void SetLevel(UShort_t val) {
        mma->level = val;
    };

    UShort_t GetOut() {
	return mma->out;
    };

    UShort_t GetLevel() {
	return mma->level;
    };

    void SetOut(UShort_t val) {
	mma->out = val;
    };

};
#endif
