#include <stdio.h>
#include <TApplication.h>
#include <TServerSocket.h>
#include <TMutex.h>
#include <poll.h>
#include <unistd.h>
#include <TMonitor.h>
#include <TObjString.h>

#include "module_numbers.h"
#include "module_ip.h"

#include "Receiver.h"
#include "GateNet.h"
#include "GathererDecoder.h"
#include "Run44Display.h"

#include "rbcp_operator.h"
#include "copper_lite.h"
#include "kalliope_tdc.h"

#include "EDB.h"

#include <sys/time.h>
#define	DO_NOT_USE_RECEIVER2

#include "niki_3n00.h"

  extern "C" {
    
#include "acs_vme.h"
#include "set_vme_module.h"
#include "read_setting.h"
    //#include "write_setting.h"
    
    int quit_vme();
    int read_address_def(struct control *c);
    int vme_init(char *hostnm,char *portnm);
    int set_a3n00_address(int module);
    int read_setting_def(struct data_file *d,struct control *cntl);
    int get_acr_state();
    int module_default_setting(int flg);
    //    int read_from_sitcp(unsigned char *buff);
    
  };
  

Long64_t FakeClock()
{
    timeval tv;
    gettimeofday(&tv, 0);

    union {
	Long64_t q64;
	UInt_t q32[2];
    } v_be;

    Long64_t v_le;

    //    Double_t t = (double)tv.tv_sec + (((double)(tv.tv_usec)+0.5) / 1000000.0);//unix-time epoch GateNet
    Double_t t = (double)(tv.tv_sec-0x7798280) + (((double)(tv.tv_usec)+0.5) / 1000000.0); //MLF-time epoch GateNet

    Long64_t sec = (long long int) t & 0x3FFFFFFF;
    Long64_t sub_sec = ((long long int)((t-int(t)) * 32000.0)) & 0x7FFF;
    Long64_t sub_usec = ((long long int)((((t-int(t)) * 32000.0)-int((t-int(t)) * 32000.0)) * 40000000.0)) & 0x7FF;

    v_le =  (0x5c00000000000000LL | (sec << 26) | (sub_sec << 11) | sub_usec);
    v_be.q32[0] = htonl(v_le >> 32);
    v_be.q32[1] = htonl(v_le & 0xFFFFFFFFL);

    return v_be.q64;
}

void
config_copper_lite(const char * host)
{
    rbcp_operator copper_lite_ctrl(host, "4660");

    /* check registers */
    copper_lite_ctrl.Read(0, 0x30);
    copper_lite_ctrl.Show();

    /* 0x80 means FIXED order, 0x80: fixed, 0x00 round robin */
    /* 0x40 means byte order, 0x00: motorola, 0x40: intel */
    /* 1, 2, 4, 8 mean slot A,B,C,D */
    unsigned char arbitration = 0x80 | 0x40 | 0xF; /* all slot, fixed order, intel byte-order */
    copper_lite_ctrl.Write(0xC, &arbitration, sizeof(arbitration));
    copper_lite_ctrl.Show();
}

void
config_kalliope_tdc(const char * host)
{
    rbcp_operator kalliope_tdc_ctrl(host, "4660");

    /* check registers */
    kalliope_tdc_ctrl.Read(0, 0x30);
    kalliope_tdc_ctrl.Show();

    /* 0x80 means FIXED order, 0x80: fixed, 0x00 round robin */
    /* 0x40 means byte order, 0x00: motorola, 0x40: intel */
    /* 1, 2, 4, 8 mean slot A,B,C,D */
    unsigned char arbitration = 0x80 | 0x40 | 0xF; /* all slot, fixed order, intel byte-order */
    kalliope_tdc_ctrl.Write(0xC, &arbitration, sizeof(arbitration));
    kalliope_tdc_ctrl.Show();
}



class Worker : public BGClass {
private:
    copper_lite		m_copper_lite[N_COPPER_LITE_MODULES];
    kalliope_tdc	m_kalliope_tdc[N_KALLIOPE_TDC_MODULES];
    kalliope_tdc	m_nim_tdc[N_NIM_TDC_MODULES];
  niki_module_reader*	m_niki_module;   //[N_NIKI_3N00_MODULES];
    GateNet		m_gatenet[N_GATENET_MODULES];
    GathererDecoder	m_decoder1;
    Run44Display	m_display;
    Run44Decoder &	m_decoder2;
    EDB			m_edb;
    TMutex		m_fileop_mutex;
    Bool_t		m_running;
    Bool_t		m_starting;
    Bool_t		m_ending;
    TDirectory *	m_rootdir;
    FILE *		m_copper_lite_dump[N_COPPER_LITE_MODULES];
    FILE *		m_kalliope_tdc_dump[N_KALLIOPE_TDC_MODULES];
    FILE *		m_nim_tdc_dump[N_NIM_TDC_MODULES];
    FILE *		m_niki_adc_dump[N_NIKI_3N00_MODULES];
    Int_t		m_dump_file_generation;
    Int_t		m_run_number;
    Long64_t		m_event_number;

public:

    Worker() :
	m_fileop_mutex(kFALSE),
	m_running(kFALSE),
	m_starting(kFALSE),
	m_ending(kFALSE),
	m_decoder2(m_display.GetDecoder()),
	m_rootdir(gDirectory),

	//	m_edb("/home/daq/Data") 
	m_edb("/data/edb") 
  {
    printf("\nEntering Worker()\n");

	m_niki_module = new niki_module_reader[N_NIKI_3N00_MODULES];

	ResetDumpFiles();

	m_decoder2.UseNewCoincidenceAlgorithm(kTRUE);

	cerr << "----------------------------" << endl;
	m_rootdir->pwd();
	cerr << "----------------------------" << endl;

	/* specify the COPPER IP addr */
	m_display.Draw("colz");
	//	printf("passed colz\n");
    printf("Exitting Worker()\n");
    };

    TMutex * FileopMutex() {
	return &m_fileop_mutex;
    };

    void
    StartCopperLite() {
      	for (int i=0; i<N_COPPER_LITE_MODULES; i++) {
	/* setup copper lite */
	config_copper_lite(copper_lite_UDP_ip[i]);
	m_copper_lite[i].Open(copper_lite_TCP_ip[i], "24");
	}
    };

    void
    StopCopperLite1() {
	for (int i=0; i<N_COPPER_LITE_MODULES; i++) {
	    cerr << "Shutdown TCP to COPPER-Lite " << i << endl;
	    m_copper_lite[i].Shutdown();
	}
    };

    void
    StopCopperLite2() {
	for (int i=0; i<N_COPPER_LITE_MODULES; i++) {
	    cerr << "closing file descriptor to COPPER-Lite " << i << endl;
	    m_copper_lite[i].Close();
	}
    };

    void
    StartKalliopeTDC() {
	for (int i=0; i<N_KALLIOPE_TDC_MODULES; i++) {
	    m_kalliope_tdc[i].SetModuleId(N_COPPER_LITE_MODULES+i);
	    cerr << "configuring " << kalliope_tdc_ip[i] << endl;
	    config_kalliope_tdc(kalliope_tdc_ip[i]);
	    cerr << "done" << endl;

//	    TThread::Sleep(1);

	    cerr << "connecting " << kalliope_tdc_ip[i] << endl;
	    m_kalliope_tdc[i].Open(kalliope_tdc_ip[i], "24");
	    cerr << "done " << kalliope_tdc_ip[i] << endl;

	}
    };

    void
    StopKalliopeTDC1() {
	for (int i=0; i<N_KALLIOPE_TDC_MODULES; i++) {
	    cerr << "Shutdown TCP to Kalliope " << i << endl;
	    m_kalliope_tdc[i].Shutdown();
	}
    };

    void
    StopKalliopeTDC2() {
	for (int i=0; i<N_KALLIOPE_TDC_MODULES; i++) {
	    cerr << "closing file descriptor to Kalliope " << i << endl;
	    m_kalliope_tdc[i].Close();
	}
    };

    void
    StartNimTDC() {
	for (int i=0; i<N_NIM_TDC_MODULES; i++) {
//	    TThread::Sleep(1);
	  m_nim_tdc[i].SetModuleId(N_COPPER_LITE_MODULES+N_KALLIOPE_TDC_MODULES+i);//ID is written in 

	cerr << "configuring NIM TDC " << nim_tdc_ip[i] << endl;
	config_kalliope_tdc(nim_tdc_ip[i]);
	cerr << "done" << endl;

	cerr << "connecting " << nim_tdc_ip[i] << endl;
	m_nim_tdc[i].Open(nim_tdc_ip[i], "24");
	cerr << "done" << endl;
	}
    };

    void
    StopNimTDC1() {
	for (int i=0; i<N_NIM_TDC_MODULES; i++) {
	cerr << "Shutdown TCP to NIM TDC" << endl;
	//	m_nim_tdc.Close();
	m_nim_tdc[i].Shutdown();
	}
    };

    void
    StopNimTDC2() {
	for (int i=0; i<N_NIM_TDC_MODULES; i++) {
	cerr << "closing file descriptor to NIM TDC" << endl;
	m_nim_tdc[i].Close();
	}
    };

    void
    StartNiki() {

      if(N_NIKI_3N00_MODULES) {

      if(read_address_def(&cntl))
	fprintf(stderr,"address error!\n");

      //      if(vme_init(hostnm,portnm)){
      if(vme_init("192.168.12.3","4660")){
	//can't connect vmebus
	fprintf(stderr,"vmebus error!\n");
	exit(0);
      }
      
      if(set_a3n00_address(0)){
	fprintf(stderr,"A3100 error!\n");
	exit(0);
      }

      if(read_setting_def(&d_f,&cntl))
	fprintf(stderr,"setting error!\n");
      
      int acr = get_acr_state();
      acr = 0;
      if(acr){
	/*now aquire */
	/*計測を止めた方がいいかも*/
	fprintf(stderr,"now pha aquire!\n");
	read_module_setting(0);// read from a3100
	/*************************/
      }else{
	module_default_setting(0); // write to a3100  setting.def
      }
      int n = cntl.modulenum;
      if(n!=N_NIKI_3N00_MODULES){
	printf("error in Niki modules numbers %d vs %d\n",n, N_NIKI_3N00_MODULES);
	exit(1);
      }
      for(int i = 0 ; i < n ; i++){
	a3n00_start_set(i);
	m_niki_module[i].module_id(i); //set module_id for hardware read
	cerr << "configuring Niki module " << i << " at VME controller " << niki_3n00_ip[i] << endl;
	//	config_kalliope_tdc(nim_tdc_ip[i]);
	cerr << "done" << endl;
	
      }
      }

    };

    void
    StopNiki1() {
      quit_vme();
    };

    void
    StopNiki2() {
    };

    void
    OpenDumpFiles(Int_t run_number, Int_t generation) {
	if (0 == access("/home/daq/raw/", R_OK|W_OK|X_OK)) {

	    TString fname;

	    /* COPPER Lite */
	    for (int i=0; i<N_COPPER_LITE_MODULES; i++) {
		fname.Form("/home/daq/raw/rawdata_run%d.copper_lite.%d.%d", run_number, i, generation);
		m_copper_lite_dump[i] = fopen(fname, "w");
	    }

	    /* Kalliope TDC */
	    for (int i=0; i<N_KALLIOPE_TDC_MODULES; i++) {
		fname.Form("/home/daq/raw/rawdata_run%d.kalliope_tdc.%d.%d", run_number, i, generation);
		m_kalliope_tdc_dump[i] = fopen(fname, "w");
	    }

	    /* NIM TDC */
	    for (int i=0; i<N_NIM_TDC_MODULES; i++) {
	      fname.Form("/home/daq/raw/rawdata_run%d.nim_tdc.%d", run_number, generation);
	      m_nim_tdc_dump[i] = fopen(fname, "w");
	    }

	    /* Niki ADC */
	    for (int i=0; i<N_NIKI_3N00_MODULES; i++) {
	      fname.Form("/home/daq/raw/rawdata_run%d.niki_adc.%d", run_number, generation);
	      m_niki_adc_dump[i] = fopen(fname, "w");
	    }
	} else {
	  ResetDumpFiles();
	}
    };

    void
    CloseDumpFiles() {
	for (int i=0; i<N_COPPER_LITE_MODULES; i++) {
	    if (m_copper_lite_dump[i])
		fclose(m_copper_lite_dump[i]);
	}
	for (int i=0; i<N_KALLIOPE_TDC_MODULES; i++) {
	    if(m_kalliope_tdc_dump[i])
		fclose(m_kalliope_tdc_dump[i]);
	}
	for (int i=0; i<N_NIM_TDC_MODULES; i++) {
	if (m_nim_tdc_dump[i])
	    fclose(m_nim_tdc_dump[i]);
	}
    };

    void ResetDumpFiles() {
	for (int i=0; i<N_COPPER_LITE_MODULES; i++) {
	    m_copper_lite_dump[i] = NULL;
	}
	for (int i=0; i<N_KALLIOPE_TDC_MODULES; i++) {
	    m_kalliope_tdc_dump[i] = NULL;
	}
	for (int i=0; i<N_NIM_TDC_MODULES; i++) {
	m_nim_tdc_dump[i] = NULL;
	}
    };

    Bool_t
    FileIsSmallerThan1G(FILE * f) const {
	if (ftell(f) > 1024 * 1024 * 1024) {
	    return kFALSE;
	} else {
	    return kTRUE;
	}
    };

    void
    RotateDumpFile() {
	// m_dump_file_generation++;
	CloseDumpFiles();
	ResetDumpFiles();
	OpenDumpFiles(m_run_number, m_dump_file_generation);
    };

    void
    NewRun(Int_t run_number) {

	/************* RUN INIT ****************/

	m_dump_file_generation = 0;

	m_event_number = 0;

	m_run_number = run_number;

	OpenDumpFiles(run_number, m_dump_file_generation);
	printf("passed OpenDumpFiles\n");
	m_edb.NewRun(run_number);
	printf("passed NewRun(run_number)\n");

	/* clear histograms, around here */
	m_display.fRunNumber = run_number;printf("passed run_number\n");
	m_display.GetDecoder().Reset();	printf("passed GetDecoder().Reset()\n");
	m_display.SetStart();           printf("passed SetStart()\n");
	m_display.fEventNumber = 0;
	m_display.fTotalEventNumber = 0;

	/* Activate VETO */
	m_display.Pause();              printf("passed Pause()\n");
	if (N_COPPER_LITE_MODULES) {
	{       /* load drivers on COPPER */
	    TSocket s1("192.168.0.99", 10999);
	    TSocket s2("192.168.0.98", 10999);
	}
	}

	StartCopperLite();    printf("passed StartCopperLite()\n");
	StartKalliopeTDC();
	StartNimTDC();
	StartNiki();

	/* beginning of the run */
	// m_gatenet.Open("127.0.0.1", 14001);
	//	if(N_GATENET_MODULES){
	for(int i=0;i<N_GATENET_MODULES;i++){
	  m_gatenet[i].Open(gatenet_ip[i]);
	}

	/* Deactivate VETO */
	m_running = kTRUE;
	m_display.SetRunning(m_running);
	m_display.Resume();
	m_ending = kFALSE;
    };

    void
    EndRun() {
	/* avoid Double END  */
      	if (m_running == kFALSE) {
	    cerr << "EndRun() is called, but not running. return" << endl;
      	    return;
	}

	m_ending = kTRUE;

	poll(0, 0, 100);

	TThread * thread = GetThread();

	StopCopperLite1();
	StopKalliopeTDC1();
	StopNiki1();
	cerr << "Done" << endl;

	cerr << "EEEEEEE" << endl;

	poll(0, 0, 100);

	if (thread) {
	    if (thread->GetState() == TThread::kRunningState) {
		cerr << "waiting receiver thread" << endl;
		thread->Join();
		cerr << "receiver thread has finished" << endl;
	    } else {
		cerr << "receiver thread is not running" << endl;
	    }
	}

	cerr << "EEEEEEEEEEEEEEEEE" << endl;

	StopCopperLite2();
	StopKalliopeTDC2();
	StopNiki2();

	cerr << "NNNNNNNN" << endl;

	m_running = kFALSE;
	m_display.SetRunning(m_running);

	cerr << "NNNNNNNNNNNNNNNNN" << endl;

	//	if(N_GATENET_MODULES){
	for(int i=0;i<N_GATENET_MODULES;i++){
	m_gatenet[i].Close();
	}
	StopNimTDC1();
	StopNimTDC2();
	cerr << "DDDDDDD" << endl;
	m_edb.EndRun();

	CloseDumpFiles();
	ResetDumpFiles();
	cerr << "DDDDDDDDDD" << endl;

	//	m_display.Update();

	cerr << "DDDDDDDDDDDD" << endl;
	m_display.UpdateNhit();
	cerr << "DDDDDDDDDDDDDDD" << endl;

	m_ending = kFALSE;
    };

    void Pause() {
	m_display.Pause();
    };

    void Resume() {
	m_display.Resume();
    };

    void
    ProcessEvent() {       //Process event for one trigger
	m_event_number ++;

	m_decoder1.Reset();
	m_decoder2.Clear(); /* dont Reset(), which means clear histograms */

	Bool_t need_to_rotate_dumpfile = kFALSE;


	/* read NIM TDC */
	if (m_event_number < 10)
	    cerr << "Reading "<<N_NIM_TDC_MODULES<<" NIM TDC" << endl;
	for (int i=0; i<N_NIM_TDC_MODULES; i++) {
	try {
	    m_nim_tdc[i].Read();
	    m_decoder1.decode_nim_tdc_event(m_nim_tdc[i].Last(), 
		 N_COPPER_LITE_MODULES+N_KALLIOPE_TDC_MODULES+i);
	    if (FILE * f = m_nim_tdc_dump[i]) {
		const kalliope_tdc_event & e = m_nim_tdc[i].Last();
		const kalliope_tdc_header & header = e.header();
		const kalliope_tdc_footer & footer = e.footer();
		fwrite(&header, sizeof(header), 1, f);
		fwrite(e.body(), e.len(), 1, f);
		fwrite(&footer, sizeof(footer), 1, f);

		if (! FileIsSmallerThan1G(f)) {
		    need_to_rotate_dumpfile = kTRUE;
		}
	    }
	} catch (char const *msg) {
	    cerr << "===========================" << endl;
	    cerr << "Reading NIM TDC ";
	    cerr << ": " << msg << endl;
	    cerr << "===========================" << endl;
	    if (m_ending == kTRUE)
		return;
	    abort();
	}
	}

	/* read copper_lite */
	if (m_event_number < 10)
	    cerr << "Reading "<<N_COPPER_LITE_MODULES<<" COPPER Lite" << endl;
	for (int cslot=0; cslot<N_COPPER_LITE_MODULES; cslot++) {
	    for (int fslot=0; fslot<4; fslot++) {
		try {
		    m_copper_lite[cslot].Read();
		    m_decoder1.decode_copper_lite_event(m_copper_lite[cslot].Last(), cslot);

		    if (FILE * f = m_copper_lite_dump[cslot]) {
			const copper_lite_event & e = m_copper_lite[cslot].Last();
			const copper_lite_header & header = e.header();
			const UInt_t & footer = e.footer();
			fwrite(&header, sizeof(header), 1, f);
			fwrite(e.body(), e.len(), 1, f);
			fwrite(&footer, sizeof(footer), 1, f);

			if (! FileIsSmallerThan1G(f)) {
			    need_to_rotate_dumpfile = kTRUE;
			}
		    }

		} catch (char const *msg) {
		    cerr << "===========================" << endl;
		    cerr << "Reading COPPER Lite " << cslot;
		    cerr << " FINESSE " << fslot << ": ";
		    cerr << msg << endl;
		    cerr << "===========================" << endl;
		    if (m_ending == kTRUE)
			return;
		    abort();
		}
	    }
	}

	/* read Kalliope TDC */

	if (m_event_number < 10)
	    cerr << "Reading "<<N_KALLIOPE_TDC_MODULES<<" Kalliope" << endl;
	for (int i=0; i<N_KALLIOPE_TDC_MODULES; i++) {
	    try {
		m_kalliope_tdc[i].Read();
		//		cerr << i;
		m_decoder1.decode_kalliope_tdc_event(m_kalliope_tdc[i].Last(), 
						     N_COPPER_LITE_MODULES+i);

		if (FILE * f = m_kalliope_tdc_dump[i]) {
		    const kalliope_tdc_event & e = m_kalliope_tdc[i].Last();
		    const kalliope_tdc_header & header = e.header();
		    const kalliope_tdc_footer & footer = e.footer();
		    fwrite(&header, sizeof(header), 1, f);
		    fwrite(e.body(), e.len(), 1, f);
		    fwrite(&footer, sizeof(footer), 1, f);

		    if (! FileIsSmallerThan1G(f)) {
			need_to_rotate_dumpfile = kTRUE;
		    }
		}

	    } catch (char const *msg) {
		cerr << "===========================" << endl;
		cerr << "Reading Kalliope " << i;
		cerr << ": " << msg << endl;
		cerr << "===========================" << endl;
		if (m_ending == kTRUE)
		    return;
		abort();
	    }
	}


	/* read NIKI 3N00 */
	if (m_event_number < 10)
	  cerr << "Reading "<<N_NIKI_3N00_MODULES<<" Niki modules" << endl;

	for (int i=0; i<N_NIKI_3N00_MODULES; i++) {
	  niki_module_reader & niki = m_niki_module[i];

	  uint32_t current_event_id = 0xFFFFFFFF;

	  if (niki.has_event()) {
	    current_event_id = niki.event().event_id();
	  }

	  while (1) {
	    //	    fprintf(stderr, "current_event_id = %d\n", current_event_id);
	    if (!niki.has_event()) {
	      if (niki.fetch_event() == 0xFFFFFFFF) {
		niki.pop();

		poll(NULL, 0, 40); /* wait for 40ms, next pulse in 25Hz */

		if (niki.fetch_event() == 0xFFFFFFFF) {
		  niki.pop();
		  fprintf(stderr, "no data has come after 40ms\n");
		  break;
		}
	      }
	    }

	    if (niki.has_event()) {
	      if (current_event_id == 0xFFFFFFFF)
		current_event_id = niki.event().event_id();

	      if (niki.event().event_id() != 0xFFFFFFFF &&
		  niki.event().event_id() != current_event_id) {
		//		fprintf(stderr, "next event %d is found in buffer\n", niki.event().event_id());
		break;
	      }

	      try {
		//		fprintf(stderr, "decode event %d in buffer\n", niki.event().event_id());
		m_decoder1.decode_niki_3n00_event(niki.event(),
					N_COPPER_LITE_MODULES+N_KALLIOPE_TDC_MODULES+N_NIM_TDC_MODULES+i);
		niki.pop();
	      } catch (char const * msg) {
		fprintf(stderr, "exception in decode_niki %s\n", msg);
		abort();
	      }
	    }
	  }
	}

	if (need_to_rotate_dumpfile)
	    RotateDumpFile();

	// m_decoder1.Truncate(200*1000); /* set upper limit of single hits */

	//	if(N_GATENET_MODULES){
	for(int i=0;i<N_GATENET_MODULES;i++){
	m_gatenet[i].RecvEvent();

	m_decoder1.Push(m_gatenet[i].GetT0());
	m_decoder1.Push(m_gatenet[i].GetClock());
	}
	if(!N_GATENET_MODULES){
	  Long64_t fake_clock_data = FakeClock();
	  	m_decoder1.Push(fake_clock_data);
		//	  m_edb.WriteData(&fake_clock_data, sizeof(fake_clock_data));
	}

	//	if (m_decoder1.GetN() < 200 * 1000) {
	m_edb.WriteData(m_decoder1.GetBuffer(), m_decoder1.GetNbytes());
	//      printf("passed WriteData\n");

	m_decoder2.decode_MLFdata(m_decoder1.GetBuffer(), m_decoder1.GetNbytes());
	//      printf("passed decode_MLFdata\n");

	m_decoder2.Coincidence(20);
	//      printf("passed Coincidence\n");
	//	}

	m_decoder2.fNevent++; /* This has been moved from Run44Display::Update() */
	if(m_decoder2.fNhit_double_per_100>100) m_decoder2.fNeventWithBeam++; 

	m_display.fEventNumber ++;
	m_display.fTotalEventNumber ++;

	if (m_display.fEventNumber == 1 || m_display.fEventNumber % 100 == 0) {
	    m_display.UpdateNhit();
	}

    };

    void
    Update() {
	if (!m_running) 
	    return;

	cerr << '_';
	m_display.Update();
	cerr << 'o';
	m_display.Dump(&m_fileop_mutex);
	cerr << 'O' << endl;
    };

    virtual void ThreadFunc() {
	while (1) {
	    if (m_ending) {
		cerr << "\nReceiver thread exits because of END RUN\n" << endl;
		return;
	    }
	    if (m_running) {
		ProcessEvent();
		cerr << '.';
	    } else {
		poll(0,0,100);
	    }
	}
    };

    Bool_t IsRunning() const {
	return m_running;
    };
};

class CanvasUpdator : public BGClass {
private:
    Worker	& m_worker;
    TStopwatch	  m_elapse;
public:
    CanvasUpdator(Worker & worker) :
	m_worker(worker) {
    };

    virtual void ThreadFunc() {
	while (1) {
	    Bool_t once_more = kFALSE;
	    if (m_worker.IsRunning() || once_more) {
		m_elapse.Start();
		cerr << "update start" << endl;
		m_worker.Update();
		cerr << "update done" << endl;
		m_elapse.Stop();
		cerr << "Took " << m_elapse.RealTime() << endl;

		if (m_worker.IsRunning()) {
		    once_more = kTRUE;
		} else {
		    once_more = kFALSE;
		}
	    }
	    TThread::Sleep(4);
	}
    };
};

class CommandReceiver : public TObject {
private:
    TServerSocket m_socket;
    Worker	& m_worker;
    TMonitor      m_monitor;
public:
    CommandReceiver(Worker & worker, Int_t port = 2222) :
	m_socket(port, /* reuse */ kTRUE ),
	m_worker(worker) {
	assert(m_socket.IsValid());
	m_monitor.Add(&m_socket);
    };

    TMonitor * GetMonitor() {
	return &m_monitor;
    };

    void
    ProcessTokenList(TObjArray * words) {

	TString cmd, arg;

	if (words->At(0)) {
	    cmd = ((TObjString *)words->At(0))->GetString();
	}

	if (words->At(1)) {
	    arg = ((TObjString *)words->At(1))->GetString();
	}

	printf("cmd=%s arg=%s\n", cmd.Data(), arg.Data());

	if (cmd.CompareTo("start") == 0) {
	    int run_number = 999999;
	    if (arg.IsDigit())
		run_number = arg.Atoi();

	    printf("start %d\n", run_number);

	    if (m_worker.IsRunning()) {
		cerr << "Already running, so new run is not started" << endl;
		return;
	    } else {
		m_worker.Start();
		m_worker.NewRun(run_number);
	    }

	} else if (cmd.CompareTo("stop") == 0) {
	    cerr << "stop command is received" << endl;

	    if (m_worker.IsRunning()) {
		cerr << "Now I'm running and call EndRun()" << endl;
		m_worker.EndRun();
		cerr << "EndRun() finished" << endl;
		m_worker.Stop();
	    } else {
		cerr << "I'm not running and ignore it" << endl;
	    }
	} else if (cmd.CompareTo("pause") == 0) {
	    printf("pause\n");
	    m_worker.Pause();
	} else if (cmd.CompareTo("resume") == 0) {
	    printf("resume\n");
	    m_worker.Resume();
	} else {
	    printf("bad_cmd\n");
	}

	cerr << "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" << endl;
    };

    void
    ProcessSocket(TSocket * s) {
	FILE * fp = fdopen(s->GetDescriptor(), "r+");

	TString str;
	str.Gets(fp);

	fprintf(fp, "\r\nyou said [[[%s]]]\r\n", str.Data());
	fflush(fp);

	TObjArray * words = str.Tokenize(" \r\n");
	words->SetOwner(kTRUE);

	ProcessTokenList(words);

	words->Delete();

	fprintf(fp, "\r\nI did [[[%s]]]\r\n", str.Data());

	fflush(fp);
	fclose(fp);
    };

    virtual void Pop() {
	TLockGuard lock(m_worker.FileopMutex());
	TSocket * s = m_socket.Accept();
	ProcessSocket(s);
	s->Close();
	s->Delete();
	cerr << "______" << endl;
    };
};

int main(int argc, char ** argv)
{
    int optchar;
    int port = 2222;


    while (-1 != (optchar = getopt(argc, argv, "p:"))) {
	switch (optchar) {
	case 'p':
	    port = strtol(optarg, 0, 0);
	    break;
	}
    }
    TApplication app(argv[0], 0, 0);
    printf("passed TApplication\n");
    Worker x;
    printf("passed worker x\n");
    CommandReceiver y(x, port);
    printf("passed CommandReceiver y\n");
    CanvasUpdator z(x);
    printf("passed CanvasUpdator z\n");

    z.Start();
    printf("passed x.Start()\n");
// y.Start();

    app.Connect(y.GetMonitor(), "Ready(TSocket *)", "TObject", &y, "Pop()");
    printf("passed app.Connect()\n");

    app.Run();
    printf("passed app.Run()\n");

}
